#--- Structures on which different model types are built
import warnings

warnings.filterwarnings("ignore", category=RuntimeWarning)

#__all__ = ['RateModel', 'FunctionModel']

class RateModel:
    """
    Model structure based on rate equations
    """
    def __init__(self, **kwargs):
        params = self.default.copy()
        params.update(kwargs)
        
        self.name = self.class_name
        
        self.keys = params.keys()
        for key, value in params.items():
            setattr(self, key, value)

        self.popnum = len(self.populations)
    
    def rate(self, N, photons):
        raise NotImplementedError
        
    def PLsig(self, N):
        raise NotImplementedError
        
    def update(self, **kwargs):
        for key, val in kwargs.items():
            if key in self.keys:
                self.__setattr__(key,val)

    def params(self):
        return {key: self.__getattribute__(key) for key in self.keys}
    
    
class FunctionModel:
    """
    Model structure where signal is a function of time (e.g. multi-exponential)
    """
    def __init__(self, **kwargs):
        params = self.default.copy()
        params.update(kwargs)
        
        self.name = self.class_name
        
        self.keys = params.keys()
        for key, value in params.items():
            setattr(self, key, value)
            
        self.populations = None
        
    def PLsig(self, t):
        raise NotImplementedError
    
    def update(self, **kwargs):
        for key, val in kwargs.items():
            if key in self.keys:
                self.__setattr__(key,val)

    def params(self):
        return {key: self.__getattribute__(key) for key in self.keys}
        


import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import numpy as np

from KinetiKit.sim.time import linear
from KinetiKit.artists import plotparams
from KinetiKit.units import units
import KinetiKit.kit as kin_kit


def plot(x, y, sys_obj=None, t_dict=None, other_params=None, ivtype='none', 
         ivkey='k_ann', unit='ns', offset=None, norm=True, 
         fig=None, annotate=True, ResetColorCyc=False, 
         opaq=1, IgnoreLabel=False, mlabel=None, color=None,
         markers=False, linewidth=None, xmin=None, xmax=None, ymin=None, ymax=None):
    """
    A versatile function for plotting series of data or simulations on three
    scale types: linear, log-linear, and log-log. It contains several arguments
    relating to simulation parameters that can help automatically annotate
    the output plot.
    
    Parameters
    ----------
    x : 1D array
        Time array of data or simulation
    y : 1D array
        Signal array of data or simulation. Must be of same size as x.
    sys_obj : sim.system object, optional
        Determines the set of parameters such as rate equation constants that 
        characterize the simulation. It is used for properly annotating the 
        graph if a simulation is displayed. Default is None.
    t_dict : dictionary, optional
        Determines the set of parameters related to time which affect the 
        simulation, e.g. `reprate`, `N`, and `subsample`. Default is
        `sim.time.linear()`. It is used to properly annotate the time 
        parameters of the simulation. In a future iteration of this function, 
        the argument `x` could become optional and be replacable by 
        `t_dict['array']` If None, then `t_dict` is later defined as 
        `sim.time.linear`.
    other_params : dictionary, optional
        If a parameter other than system or time parameters is to be varied 
        (e.g. emission color for a single system, laser power, etc.), this
        dictionary should contain the keys and values for the additional 
        parameters that are relevant.
    ivtype : a string out of `'eq'`, `'tm'`, `'other'`, and `'none'`, optional
        Specifies the type of parameter that is varied across different traces
        that are plotted. `'eq'` specifies a simulation parameter, `'tm'` a time 
        parameter, `'other'` an additional parameter, and `'none'` the lack of 
        any varied parameter. Default is 'none'. 
    ivkey : string representing an existing dictionary key, optional
        If `ivtype` is 'eq', 'tm', or 'other', ivkey specifies the key of the 
        parameter that is varied in the system, time, or other parameter 
        dictionary, respectively. Default is None. If any parameter is to be 
        varied, `ivtype` and `ivkey` must always be defined.
    unit : string representing key of ``TRPL.units.units`` dictionary, optional
        Determines the unit written on x-axis and the value by which to divide 
        time array. Default is 'ns'.
    offset : array or length of length 3, optional
        If desired, plots can be made with a time offset specified in
        appropriate time units. The values of `offset[0]`, `offset[1]`, and
        `offset[2]` determine the shifts in the linear, log-linear, and log-log
        plots, respectively. Default is None.
    norm : Boolean, optional
        Specifies whether normalized or unnormalized graphs should be plotted. 
        Default is True.
    fig : Figure object or None
        Determines whether the plot traces should be made on a new figure or on
        an existing set of axes. Default is None, in which case a new figure 
        is plotted. If a figure object is provided, it must have (1, 3) 
        subplots
    annotate : Boolean, optional
        Specifies whether to display the simulation and time parameters at the
        bottom of the three graphs. Default is True. The size of the figure 
        depends on the existence or absence of annotation.
    ResetColorCyc : Boolean, optional
        If True, the color cycler will be reset to the first value specified 
        in ``artists.plotparams``. Should be used in conjunction with varying
        opaq so that traces with similar color can always be distinguished. 
        Can be useful when plotting two traces (e.g. emission color of the 
        same system) and want to show how each of them varies with a given 
        parameters. See ``simRates_hetero_vary.py`` for an example.
    opaq : Float between 0 and 1, optional
        The `alpha` value for trace (same across all 3 plots). Default is 1.
    IgnoreLabel : Boolean, optional
        Determines whether to omit the label of a particular trace from the 
        legend. Default is False.
    mlabel : string or None
        Stands for 'manual label'; can be used to create a label for a given 
        trace that is independent of the simulation or time parameters.
    color : string representing a color
        If not None, this setting will override the color set by the color cycle.
    markers : Boolean
        If True, the trace is plotted as markers instead of a continuous line.
    linewidth : float or None
        Default None. If specified, defines the line width of the plotted trace.
    xmin : float or None
        Default is`None`, which uses the lower limit of the time axis `x`. Defines
        the lower x-limit of the plot.
    xmax : float or None
        Analogous to xmin but for upper x limit of the plot.
    ymin : float or None
        Determines the lower y-limit of the plot. If `None`, ymin defaults to 
        0.001 if `norm=True`, or takes the y-limits of the input array `y`, if
        `norm=False`.
    ymax : float or None
        Analogous to ymin. If `None` and `norm=True`, the y-limit will be 1.20.
    
    Returns
    -------
    A figure (or a plot on an existing figure) containing a set of plots 
    represented on a linear, log-linear, and log-log scale.
    """
    
    if sys_obj is None:
        system_params = {}
    else:
        system_params = sys_obj.params().copy()
    
    if t_dict is None:
        t_dict = linear()
    else: 
        t_dict = t_dict.copy()
        
    time_params = {}
    for key, val in t_dict.items():
        if key == 'reprate' or key == 'N' or key == 'subsample' or key=='numcycles':
            time_params[key] = val
    
    varname = ivkey

    # preparing strings that will appear on figure legend and text
    if ivtype == 'tm':
        var = time_params.pop(ivkey)

        if varname == 'reprate':
            varname = 'reprate (MHz)'
            var_label = '%i' % var/1e6

        if varname == 'N' and var >= 1000:
            var_label = '%0.1fK' % (var/1000)
        
        
        else:
            var_label = '%i' % var

        rest_time_params = time_params
        rest_system_params = system_params

    elif ivtype == 'eq':
        var = system_params.pop(ivkey)
        var_label = '%0.2e' % var

        rest_time_params = time_params
        rest_system_params = system_params

    elif ivtype == 'other':
        var = other_params[varname]
        
        if varname == 'power':
            varname = 'Power (μW)'
            var_label = '%0.3f' % (var / 1e-6)
        
        else:
            var_label = '{}'.format(var)

        rest_time_params = time_params
        rest_system_params = system_params
    
    elif ivtype == 'none':
        varname = None
        var_label = None
        
        rest_time_params = time_params
        rest_system_params = system_params
    
    else:
        print("Invalid parameter type; must be 'tm', 'eq', 'other', or 'none' string.")
        return 0
    
    #--- Specifying unit
    unit_name = unit
    unit_value = units[unit]
    
    if annotate:
        txt = []
        for key, val in rest_time_params.items():
            if key == 'reprate':
                val /= 1e6
                p = 'reprate (MHz) = {0}'.format(int(val))

            elif key == 'N':
                val /= 1000
                val = int(val)
                p = ' {0} = {1}K'.format(key, val)
                
            else:
                p = '{0} = {1}'.format(key, val)
            
            txt.append(p)
        
        nL = True  # new line that separates time_params from system_params display
        
        for key, val in rest_system_params.items():
            p = '{0} = {1:0.2e}'.format(key, val)
            if nL:
                txt.append('\n'+p)
                nL = False
            else:
                txt.append(p)
        
        #print(txt)
    
    # Figure parameters
    fw = 13; fh = 5
    
    if fig is None:
        fresh_fig, (ax1, ax2, ax3) = plt.subplots(1,3, figsize=(fw, fh))

    else:
        fresh_fig = fig
        (ax1, ax2, ax3) = fig.axes
    
    if offset is None:
        offset = [0, 0 , 0 ]

    for i, ax in enumerate([ax1, ax2, ax3]):
        
        if ResetColorCyc:
            ax.set_prop_cycle(None)
            
        if mlabel is None:
            label = var_label
            leg_title = varname
        
        else:
            label = mlabel
            leg_title = None
            
        if IgnoreLabel:
            label = None
        
        xoff = x + offset[i]
        xoff /= unit_value

        if not norm:
            if markers:
                ax.plot(xoff, y, 'o', ms=4, 
                        mfc=ax.get_facecolor(), 
                        color=color, label=label, 
                        alpha=opaq)
            else:
                ax.plot(xoff, y, color=color, label=label, alpha=opaq,
                        linewidth=linewidth)

        else:
            if markers:
                ax.plot(xoff, kin_kit.normalized(y), 'o', ms=4, 
                        mfc=ax.get_facecolor(),
                        color=color, label=label, alpha=opaq)
            else:
                ax.plot(xoff, kin_kit.normalized(y), 
                    color=color, label=label, alpha=opaq, linewidth=linewidth)

        if fig is None:
            ax.set_xlabel('Time (%s)'%unit_name)
            if norm:
                ax.set_ylabel('Rel. Intensity')
            else:
                ax.set_ylabel('Counts')
                
        leg = ax1.legend(title=leg_title, fontsize=8, loc=1)
        plt.setp(leg.get_title(), fontsize=10)
    
    if xmin is None:
        xmin = 0.1
    if xmax is None:
        xmax = max(x)/unit_value
    
    ax1.set_xlim(0, xmax)
    ax2.set_xlim(0, xmax) 
    ax3.set_xlim(xmin, xmax)
    
    if norm:
        if ymax is None:
            ymax = 1.20
        if ymin is None:
            ymin = 0.001
            
        ax1.set_ylim(-0.05, 1.05)
        ax2.set_ylim(ymin, ymax) #0.001
        ax3.set_ylim(ymin, ymax)
    
    else:
        # set default ymax, ymin according to data limits
        if ymax is None: 
            # check if previously plotted trace has a higher maximum; 
            # if not, replace the current ymax with max(y)
            ymax = max([max(y), find_max_value(ax2)[1]])
        if ymin is None:                
            # analogous check for ymin
            ymin = min([min(y), find_min_value(ax2)[1]])
            
        ax1.set_ylim(-0.05*(ymax-ymin), 1.05*ymax)
        ax2.set_ylim(0.1*ymin, 2*ymax)
        ax3.set_ylim(0.1*ymin, 2*ymax)
    
    ax2.set_yscale('log')
    ax3.set_xscale('log')
    ax3.set_yscale('log')
    
    if fig is None:
        fresh_fig.tight_layout()
    
    if annotate:
        fresh_fig.text(0, -0.2, ' , '.join(txt), transform = ax1.transAxes,
                 fontsize=8, color='darkgrey', fontstyle='italic')
        fresh_fig.subplots_adjust(bottom=0.202)
    
    return fresh_fig
        
def show_species(time, species_sets, system, norm, save=False, filename=None, destfolder=None):
    """
    Application of the plot3scales.plot function targeted at plotting species
    populations
    
    """
    species_names = system.populations
    alphas = np.linspace(0.3, 1, num=len(species_sets))
    if len(alphas)==1:
        alphas[0] = 1
    fig=None
    for k, species_set in enumerate(species_sets):
        for i, species in enumerate(species_set):
            species = kin_kit.align_by_max(species, time, species_sets[0][0], avgnum=5, value=0.5*units['ns'])
            fig=plot(time, species, sys_obj=None, t_dict=None,
                     annotate=False, fig=fig, linewidth=2, 
                     ResetColorCyc=i==0, opaq = alphas[k], norm=norm,
                     mlabel=species_names[i])
    plt.tight_layout()
    plt.show()
    
    if save:
        kin_kit.save_species(filename, destfolder, norm=norm)
    
    return fig

def find_min_value(axis, which='y'):
    """
    Finds and returns the minimum y value among all plotted traces in an axis
    object.
    
    
    Parameters:
    -----------
    axis : axis object
        The axis (plot) of which we are trying to obtain the minimum value
    which : string, 'x' or 'y', optional
        Whether to look for the minimum x or y value. Default is y
    
    Returns:
    --------
    datapoint : tuple
        The x and y value of the found datapoint. Warning: if the minimum y 
        value is sought but there are multiple occurences of the same minimum 
        value, this function will return an x,y tuple where the x value 
        corresponds to the FIRST occurence of the minimum y value.
    
    """
    global_min = np.inf
    global_idx = None
    global_line = None
    
    lines = axis.lines
    
    if len(lines) == 0:
        return (np.inf, np.inf)
    
    for line in lines:
        # Step 1: find the local minimum of each line
        if which == 'y':
            local_min = line.get_ydata().min()
            local_idx = np.argmin(line.get_ydata())
        else:
            local_min = line.get_xdata().min()
            local_idx = np.argmin(line.get_xdata())
        # Step 2: check it against the current global minimum
        if local_min < global_min:
            global_min = local_min
            global_idx = local_idx
            global_line = line
        # Step 3: repeat for all lines
    
    # Step 4: Generate tuple
    if which == 'y':
        x_val = global_line.get_xdata()[global_idx]
        y_val = global_min
    else:
        x_val = global_min
        y_val = global_line.get_ydata()[global_idx]
    return x_val, y_val

def find_max_value(axis, which='y'):
    """
    Exact analogue of ``find_min_value()`` function, but for obtaining the 
    global maximum of the plots in a given axis object.
    
    """
    global_max = -np.inf
    global_idx = None
    global_line = None
    
    lines = axis.lines
    
    if len(lines) == 0:
        return (-np.inf, -np.inf)
    
    for line in lines:
        # Step 1: find the local minimum of each line
        if which == 'y':
            local_max = line.get_ydata().max()
            local_idx = np.argmax(line.get_ydata())
        else:
            local_max = line.get_xdata().max()
            local_idx = np.argmax(line.get_xdata())
        # Step 2: check it against the current global minimum
        if local_max > global_max:
            global_max = local_max
            global_idx = local_idx
            global_line = line
        # Step 3: repeat for all lines
    
    # Step 4: Generate tuple
    if which == 'y':
        x_val = global_line.get_xdata()[global_idx]
        y_val = global_max
    else:
        x_val = global_max
        y_val = global_line.get_ydata()[global_idx]
    return x_val, y_val
    



# -*- coding: utf-8 -*-
"""
Functions that output an overlay of simulation and data given a system, 
a set of y data, and parameters. 

@author: Natalia Spitha
"""

import numpy as np
from ipywidgets import interact, fixed
import ipywidgets as widgets

from KinetiKit import kit as kin_kit
from KinetiKit.units import units, ns

def active_widgets(system, dictionary, total_args=20):
    """
    This function:
        - accepts a dictionary of widgets each corresponding to a 
        system parameters
        - ensures the dictionary has the parameters in the correct order, 
        i.e. the order defined in sim.systems.ChosenSystem.
        - returns a list of widgets, with size total_args. This list contains
        the widgets for the system parameters to be displayed interactively, 
        and contains inactive (fixed(0)) widgets for the rest of the slots 
        in the list.
        
    """
    
    if system.params().keys() != dictionary.keys():
        print('Ensure that \'parameters\' contains ALL parameters of the system.')
        return
    
    else:
        ordered_dict = system.default.copy()
        ordered_dict.update(dictionary)
        dictionary = ordered_dict
        
        all_widgets = list(np.zeros(total_args))
        param_list = list(dictionary.keys())
        for i, widg in enumerate(all_widgets):
            if i < len(param_list):
                all_widgets[i] = dictionary[param_list[i]]
            else:
                all_widgets[i] = fixed(0)
        # all_widgets[np.index(all_widgets==0] = fixed(0)
        return all_widgets

def powerwidget(slidepower=False, pmin=0.001, pmax=1.5, pset = 0.5, step=0.01, power=0, label='power', unit='μW'):
    if not slidepower:
        return fixed(power)
    else:
        return widgets.FloatSlider(value=pset, min=pmin, max=pmax, step=step, description = '%s (%s)'%(label,unit),
                                   disabled=not slidepower, continuous_update=False, orientation='Horizontal',
                                   readout=slidepower, readoutformat='.3f')

def makeWidget(key, val, slider='linear', min_val=None, max_val=None, step=0.01, disabled=False,
               continuous_update=False, orientation='horizontal', readout=None,
               readout_format=None):
    
    #slide should be 'linear', 'log', or 'fixed'
    if readout is None:
        readout = not disabled
    
    if slider == 'linear':
        sliderfunc = widgets.FloatSlider
        if min_val is None:
            min_val = 0
        if max_val is None:
            max_val = 2*val
        if readout_format is None:
            precision = kin_kit.precision(step)
            readout_format = '.{pr}f'.format(pr=precision)
        if disabled:
            precision = kin_kit.precision(val)
            readout_format = '.{pr}f'.format(pr=precision)
            
            
    elif slider == 'log':
        sliderfunc = widgets.FloatLogSlider
        # FloatLogSlider accepts the actual value as `value`, but accepts 
        # log limits instead of actual power limits. So, the min_val and
        # max_val provided are converted into their log values before the
        # slider function is called.
        
        if min_val is None:
            min_val = np.log10(val)-3
        else: 
            min_val = np.log10(min_val) 
            
        if max_val is None:
            max_val = np.log10(val)+3
        else:
            max_val = np.log10(max_val)
        
        if readout_format is None:
            readout_format = '0.3e'
            
    elif slider == 'fixed':
        return fixed(val)
    
    return sliderfunc(value=val,
                          min=min_val,
                          max=max_val,
                          step=step,
                          description=key, 
                          disabled=disabled,
                          continuous_update=continuous_update,
                          orientation=orientation,
                          readout=readout,
                          readout_format=readout_format)
    
        


from KinetiKit.units import ns, ps, fs, MHz
import numpy as np


def linear(period=12.5 * ns,N=5000, subsample=1):
    """Time parameters for simulation.
    Assumes linearly spaced datapoints.

    Optional Parameters
    ----------
    period : float
        Duration of the time axis, in units of seconds.  For pulsed
        simulations, repetition rate is `1 / period`.  Default value is
        12.5 ns.
    N : int
        Number of time points sampled across `period`. Default is 50000.
    subsample : int
        How many time points are simulated between recorded data points.
        Default is 20. Simulation output will have `N // subsample` time
        points.
        
    Returns
    -------
    dictionary
        A collection of parameters relating to numerical integration and
        simulation output.
    """
    dic = {
        'period': period,
        'N': N,
        'subsample': subsample
    }

    dic['dt'] = period / N
    dic['stepsize'] = dic['dt'] * subsample
    dic['array'] = np.linspace(0, period, N + 1)[:-1]

    return dic

def get_linear_args(timeobject):
    timekeys = ['period', 'N', 'subsample']
    linear_args = {}
    for key, val in timeobject.items():
        if key in timekeys:
            linear_args[key] = val
    return linear_args

def update_linear(timeobject, **kwargs):
    timekeys = ['period', 'N', 'subsample']
    new_args = get_linear_args(timeobject)
    for key, val in kwargs.items():
        if key in timekeys:
            new_args[key] = val
        else:
            print("Expected one of the following: \n", timekeys)
    return linear(**new_args)


"""
Snippet to calculate computation time
Copy and paste:
    
time_start = time.clock()

at start of process and:
    
time_elapsed = (time.clock() - time_start)

at end of program.
"""


import numpy as np
from scipy.signal import convolve, savgol_filter
from matplotlib import pyplot as plt

from KinetiKit.units import units, MHz, fs, nm, uW, ps
import KinetiKit.kit as kin_kit

__all__ = ['convolve_irf', 'build_irf']

# --- IRF Functions
def convolve_irf(pl, t, args):
    """
    Convolves simulated function with a specified Instrument Response Function.
    IRF can be either simulated as a simple Gaussian or as an exponentially 
    modified Gaussian with a diffusion tail. See ``build_irf`` for the default 
    values for IRF construction.
    
    Parameters
    ----------
    pl : numpy array
        The signal to be convolved must be a 1-D or 2-D array, with the lowest
        dimension having the same length as the time array`t`
    t : numpy array
        1D array representing the time axis.
    args : dictionary
        Arguments to be input into `build_irf` function. See definition of
        ``build_irf`` for default values.
    
    Returns
    -------
    An array of the shape of `pl` representing the signal convolved with the 
    IRF.
    """
    
    irf = build_irf(t, **args)
    if pl.ndim == 2:
        irf = irf[np.newaxis]
    return convolve_cyclic_boundary(pl, irf)


def build_irf(t, irf_type='Gauss', weighted=True, fwhm=55 * ps, tau_wt=40 * ps, 
              tau=650 * ps, b=0.13, c=0, return_x = False):
    """
    Constructs a Gaussian-like curve simulating an instrument response function
    of a time-resolved measurement system. 
    
    Parameters
    ----------
    t : numpy array
        The time array on which to build the IRF
    irf_type : string
        Type of IRF to be constructed; ```irf_type = 'Gauss'``` will return a 
        normal Gaussian curve, while ```irf_type = 'GaussDiff'``` will return
        an exponentially modified Gaussian with a diffusion tail.
    weighted : Boolean
        Whether to use a pure Gaussian or exponentially weighted Gaussian for 
        the main "peak" of the IRF. Default is True. 
        See en.wikipedia.org/wiki/Exponentially_modified_Gaussian_distribution. 
    fwhm : float
        The FWHM that determines the variance of either type of curve. Must 
        have the same time units as `t`.
    tau_wt : float
        lifetime of decay of exponentially weighted Gaussian. Ignored if 
        `weighted` is set to False.
    tau : float
        lifetime of diffusion tail in the time units of `t`. Ignored if 
        `irf_type` is set to `'Gauss'`.
    b : float
        Relative contribution of diffusion tail to IRF. Ignored if 
        `irf_type` is set to `'Gauss'`.
    c : float
        Offset for IRF, recommended as zero.

    """
    
    
    
    if irf_type == 'Gauss':
        tirf = t[t <= 3 * fwhm]
        tirf -= tirf.mean()
        irf = kin_kit.Gauss(tirf, 1, 0, fwhm)
    elif irf_type == 'GaussDiff':
        tirf = t[t <= 100 * fwhm]
        irf = kin_kit.GaussDiff(tirf, (np.max(tirf)-np.min(tirf))/50+np.min(tirf), 1, fwhm=fwhm, tau=tau, 
                                b=b, c=c, weighted=weighted, tau_wt=tau_wt)
        window_length = int(0.006*len(t))
        if window_length%2 == 0:
            window_length += 1
        irf = savgol_filter(irf, window_length=window_length, polyorder=1)
    if return_x:
        return tirf, irf/irf.sum()
    else:
        return irf / irf.sum()

def convolve_cyclic_boundary(data, irf):
    m = irf.size
    out = np.concatenate(
        (data[..., -m:], data[..., :], data[..., :m]), axis=-1)
    out = convolve(out, irf, mode='same')
    return out[..., m:-m]



# -*- coding: utf-8 -*-
"""
Functions that output an overlay of simulation and data given a system, 
a set of y data, and parameters. 

@author: Natalia Spitha
"""

import numpy as np

from KinetiKit import sim
from KinetiKit import artists as art
from KinetiKit.artists.plotparams import colors as color_range
from KinetiKit import kit as kin_kit
from KinetiKit.units import units, ns, ps


def MonoViz(system, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14,
        p15, p16, p17, p18, p19, p20,
        to=sim.time.linear(), N_coarse=500, power = 1e-6, irf_args = {},
        data=None, power_unit='microWatt', 
        align_by = 'steep', avgnum= 5, xmin=0.1, xmax=None, 
        ymin=1e-3, ymax=1.2):
    
    args = p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14,\
        p15, p16, p17, p18, p19, p20,
    param_names = system.params().keys()
    trunc_args = args[:len(list(param_names))]
    params = kin_kit.dict_from_list(trunc_args, param_names)
    system.update(**params) # system is updated according to the parameters provided as *args
           
    #--- Creating Time Array
    dtime = to['array'][::to['subsample']]
    
    if isinstance(power, np.float) or isinstance(power, np.int):
        power_list = [power]
        power *= units[power_unit]
        light = sim.lib.Excitation(pulse={'power' : power})
        transient, converged = sim.lib.refined_simulation(system, to, light,
                                                  N_coarse=N_coarse)
        pl = system.PLsig(transient)
    
    elif isinstance(power, list):
        power_list = power
        for i, power in enumerate(power):
            power *= units[power_unit]
            light = sim.lib.Excitation(pulse={'power' : power})
            transient, converged = sim.lib.refined_simulation(system, to, light,
                                                      N_coarse=N_coarse)
            pl_at_this_power = system.PLsig(transient)
            
            if i == 0:
                pl = pl_at_this_power
            else:
                pl = np.vstack((pl, pl_at_this_power))
                
    sims = sim.lib.convolve_irf(pl, dtime, irf_args)   
    # Aligns data with sim either by max. or steep
    if align_by == 'steep':
        aligned_sims = kin_kit.align_by_steep(sims, dtime, value = 0.5*ns, avgnum=avgnum)
    
    elif align_by == 'max':
        aligned_sims = kin_kit.align_by_max(sims, dtime, value = 0.5*ns, avgnum=avgnum)
    else:
        print('\'align_by\' must be \'steep\' or \'max\'')
    aligned_sims = kin_kit.make_2d(aligned_sims)   
    
    fig=None
    data_ended = False; sims_ended = False
    for i in range(20):
        if data is None:
            pass
        else:
            data=kin_kit.make_2d(data)
            try:
                d = data[i]
                fig=art.plot3scales.plot(dtime, d, sys_obj=None, t_dict=None,
                         annotate=False, fig=fig, linewidth=1, 
                         color=color_range[i],
                         mlabel='data_%0.3f'%(power_list[i]),
                         xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax)
                         
            except IndexError:
                data_ended = True
                pass
                
            try:
                aligned_sim = aligned_sims[i]
                art.plot3scales.plot(dtime, aligned_sim, system, t_dict=to, ivtype='none', 
                                 annotate=True, fig=fig, ResetColorCyc=i==0, 
                                 color=color_range[i],
                                 linewidth=8, opaq=0.4,
                                 mlabel='sim_%0.3f'%(power_list[i]),
                                 xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax)
                                 
            except IndexError:
                sims_ended = True
                pass
            
            if sims_ended and data_ended:
                break
            
    return

def HeteroViz(system, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14,
        p15, p16, p17, p18, p19, p20,
        to=sim.time.linear(), N_coarse=500, power = 1e-6, irf_args = {}, 
        data=None, power_unit='microWatt', ids = ['layer 1', 'layer 2'],
        light=sim.lib.Excitation(),
        align_by = 'steep', avgnum= 5, xmin=0.1, xmax=None, ymin=1e-3, ymax=1.2):
    
    args = p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14,\
        p15, p16, p17, p18, p19, p20,
    param_names = system.params().keys()
    trunc_args = args[:len(list(param_names))]
    params = kin_kit.dict_from_list(trunc_args, param_names)
    system.update(**params) # system is updated according to the parameters provided as *args    
    #--- Creating Time Array
    dtime = to['array'][::to['subsample']]
    
    power *= units[power_unit]
    light = light.updated_with(pulse={'power' : power})
    transient, converged = sim.lib.refined_simulation(system, to, light,
                                                  N_coarse=N_coarse)
    pl = kin_kit.make_2d(system.PLsig(transient))
    sims = sim.lib.convolve_irf(pl, dtime, irf_args)   
    
    # Aligns data with sim either by max. or steep
    if align_by == 'steep':
        aligned_sims = kin_kit.align_by_steep(sims, dtime, value = 0.5*ns, avgnum=avgnum)
    
    elif align_by == 'max':
        aligned_sims = kin_kit.align_by_max(sims, dtime, value = 0.5*ns, avgnum=avgnum)
    else:
        print('\'align_by\' must be \'steep\' or \'max\'')
    
    aligned_sims = kin_kit.make_2d(aligned_sims)   
    
    fig=None
    
    data_ended = False; sims_ended = False
    for i in range(20):
        if data is None:
            pass
        else:
            data=kin_kit.make_2d(data)
            try:
                d = data[i]
                fig=art.plot3scales.plot(dtime, d, sys_obj=None, t_dict=None,
                         annotate=False, fig=fig, linewidth=1, 
                         color=color_range[i],
                         mlabel='data_%s'%(ids[i]),
                         xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax)
            except IndexError:
                data_ended = True
                pass
                
            try:
                aligned_sim = aligned_sims[i]
                art.plot3scales.plot(dtime, aligned_sim, system, t_dict=to, ivtype='none', 
                                 annotate=True, fig=fig, ResetColorCyc=i==0, 
                                 color=color_range[i],
                                 linewidth=8, opaq=0.4, 
                                 mlabel='sim_%s'%(ids[i]),
                                 xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax)
                
            except IndexError:
                sims_ended = True
                pass
            
            if sims_ended and data_ended:
                break
            
    return

def MultiPowerViz(system, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14,
        p15, p16, p17, p18, p19, p20,
        to=sim.time.linear(), refined=True, N_coarse=500, pulse_power = 1, cw_power=0, which=None,
        irf_args = {'fwhm':55*ps}, data=None, ids = ['one', 'two'], power_unit='microWatt', light=sim.lib.Excitation(),
        align_by = 'steep', avgnum= 5, xmin=0.1, xmax=None, ymin=1e-3, ymax=1.2):
    """
    more functionality to allow for CW and pulsed power variations
    """
    args = p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14,\
        p15, p16, p17, p18, p19, p20,
    param_names = system.params().keys()
    trunc_args = args[:len(list(param_names))]
    params = kin_kit.dict_from_list(trunc_args, param_names)
    system.update(**params) # system is updated according to the parameters provided as *args
    
    #--- Creating Time Array
    dtime = to['array'][::to['subsample']]
    
    if which is None:
        pulse_power *= units[power_unit]
        cw_power *= units[power_unit]
        light = light.updated_with(pulse={'power' : pulse_power}, cw = {'power': cw_power})

        if refined: 
            transient, converged = sim.lib.refined_simulation(system, to, light,
                                                      N_coarse=N_coarse)
        else:
            transient, converged = sim.lib.simulate_until_steady(system, to, light)
        
        pl = kin_kit.make_2d(system.PLsig(transient))
        sims = sim.lib.convolve_irf(pl, dtime, args=irf_args)   
        
        # Aligns data with sim either by max. or steep
        if align_by == 'steep':
            aligned_sims = kin_kit.align_by_steep(sims, dtime, value = 0.5*ns, avgnum=avgnum)
        
        elif align_by == 'max':
            aligned_sims = kin_kit.align_by_max(sims, dtime, value = 0.5*ns, avgnum=avgnum)
        else:
            print('\'align_by\' must be \'steep\' or \'max\'')
        
        aligned_sims = kin_kit.make_2d(aligned_sims)   
        
        fig=None
        
        data_ended = False; sims_ended = False
        for i in range(20):
            if data is None:
                pass
            else:
                data=kin_kit.make_2d(data)
                try:
                    d = data[i]
                    fig=art.plot3scales.plot(dtime, d, sys_obj=None, t_dict=None,
                             annotate=False, fig=fig, linewidth=1, 
                             color=color_range[i],
                             mlabel='data_%s'%(ids[i]),
                             xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax)
                except IndexError:
                    data_ended = True
                    pass
                    
                try:
                    aligned_sim = aligned_sims[i]
                    art.plot3scales.plot(dtime, aligned_sim, system, t_dict=to, ivtype='none', 
                                     annotate=True, fig=fig, ResetColorCyc=i==0, 
                                     color=color_range[i],
                                     linewidth=8, opaq=0.4, 
                                     mlabel='sim_%s'%(ids[i]),
                                     xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax)
                    
                except IndexError:
                    sims_ended = True
                    pass
                
                if sims_ended and data_ended:
                    break
        return
    
    else:
        fig=None
        data_ended = False
        for i in range(20):
            if data is None:
                pass
            else:
                data=kin_kit.make_2d(data)
                try:
                    d = data[i]
                    fig=art.plot3scales.plot(dtime, d, sys_obj=None, t_dict=None,
                             annotate=False, fig=fig, linewidth=1, 
                             color=color_range[i],
                             mlabel='data_%s'%(ids[i]),
                             xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax)
                except IndexError:
                    data_ended = True
                    pass
                
                if data_ended:
                    break
        
        if which == 'pulse':
            var_power = np.array(pulse_power) * units[power_unit]
            set_power = cw_power * units[power_unit]

            
        if which == 'cw':
            var_power = np.array(cw_power) * units[power_unit]
            set_power = pulse_power * units[power_unit]
        
        for p, power in enumerate(var_power):
            if which == 'pulse':
                light = light.updated_with(pulse = {'power':power}, cw = {'power':set_power})
            elif which == 'cw':
                light = light.updated_with(pulse={'power': set_power}, cw = {'power': power})
            else:
                print(' "which" should be "cw" or "pulse" ')
            #print('cw_power ', light.cw_power)    
            if refined: 
                transient, converged = sim.lib.refined_simulation(system, to, light,
                                                          N_coarse=N_coarse)
            else:
                transient, converged = sim.lib.simulate_until_steady(system, to, light)
            
            pl = system.PLsig(transient)
            sims = sim.lib.convolve_irf(pl, dtime, args=irf_args)   
            
            # Aligns data with sim either by max. or steep
            if align_by == 'steep':
                aligned_sims = kin_kit.align_by_steep(sims, dtime, value = 0.5*ns, avgnum=avgnum)
            
            elif align_by == 'max':
                aligned_sims = kin_kit.align_by_max(sims, dtime, value = 0.5*ns, avgnum=avgnum)
            else:
                print('\'align_by\' must be \'steep\' or \'max\'')
                        
            art.plot3scales.plot(dtime, aligned_sims, system, t_dict=to, ivtype='none', 
                                     annotate=True, fig=fig, ResetColorCyc=i==0, 
                                     color=color_range[p],
                                     linewidth=8, opaq=0.4, 
                                     mlabel='sim_%s'%(ids[p]),
                                     xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax)
        
    return
    
def FuncViz(system, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14,
        p15, p16, p17, p18, p19, p20,
        to=sim.time.linear(), N_coarse=500, power = 1e-6, irf_args={},
        data=None, power_list=[1], power_unit='microWatt', 
        align_by = 'steep', avgnum= 5, slidepower=False,
        xmin=0.1, xmax=None, ymin=1e-3, ymax=1.2):
    
    args = p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14,\
        p15, p16, p17, p18, p19, p20,
    param_names = system.params().keys()
    trunc_args = args[:len(list(param_names))]
    params = kin_kit.dict_from_list(trunc_args, param_names)
    system.update(**params) # system is updated according to the parameters provided as *args
    
            
    #--- Creating Time Array
    dtime = to['array'][::to['subsample']]
    
    pl = system.PLsig(dtime)
    
    sims = sim.lib.convolve_irf(pl, dtime, irf_args)  
    
    # Aligns data with sim either by max. or steep
    if align_by == 'steep':
        aligned_sims = kin_kit.align_by_steep(sims, dtime, value = 0.5*ns, avgnum=avgnum)
    
    elif align_by == 'max':
        aligned_sims = kin_kit.align_by_max(sims, dtime, value = 0.5*ns, avgnum=avgnum)
    else:
        print('\'align_by\' must be \'steep\' or \'max\'')
    aligned_sims = kin_kit.make_2d(aligned_sims)   
    
    fig=None
    data_ended = False; sims_ended = False
    for i in range(20):
        if data is None:
            pass
        else:
            data=kin_kit.make_2d(data)
            try:
                d = data[i]
                fig=art.plot3scales.plot(dtime, d, sys_obj=None, t_dict=None,
                         annotate=False, fig=fig, linewidth=1, 
                         color=color_range[i],
                         mlabel='data_%0.3f'%(power_list[i]),
                         xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax
                         )
            except IndexError:
                data_ended = True
                pass
                
            try:
                aligned_sim = aligned_sims[i]
                art.plot3scales.plot(dtime, aligned_sim, system, t_dict=to, ivtype='none', 
                                 annotate=True, fig=fig, ResetColorCyc=i==0, 
                                 color=color_range[i],
                                 linewidth=8, opaq=0.4,
                                 mlabel='sim_%0.3f'%(power_list[i]),
                                 xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax
                                 )
            except IndexError:
                sims_ended = True
                pass
            
            if sims_ended and data_ended:
                break
            
    return



"""
Simple mathematical methods used throughout the module
"""

import sys
import os
import csv

import numpy as np
from scipy import special
import matplotlib.pyplot as plt

def normalized(ar, alert=True):
    """
    Returns an array normalized by its maximum value.
    
    Assumes that ar is to be normalized by a positive value.
    
    Parameters
    ----------
    ar : arr
        The array to be normalized
    
    alert : boolean, optional
        Determines whether or not to print a warning if the maximum of 
        ar_max is less than or equal to zero. Default is True.
    
    Returns
    -------
    norm_ar : arr
        Normalized array
    """
    norm_ar = ar.copy()
    
    if ar.ndim > 2:
        print('Error: Normalized function accepts only 2D or 1D arrays')
        #issue_error
        return
    
    if ar.ndim == 1:
        norm_ar = np.reshape(norm_ar, (1, len(ar)))
        
    for i, raw_ar in enumerate(norm_ar):
        raw_ar_max = raw_ar.max()
        if raw_ar_max <= 0:
            norm_ar[i] = raw_ar
            if alert:
                print('Array does not contain non-zero positive values!')
        else:  
            norm_ar[i] = raw_ar / raw_ar_max
            
    if ar.ndim == 1:
        return norm_ar[0]
    else:
        return norm_ar

def precision(number):
    """
    Returns the last decimal position of the number

    Parameters
    ----------
    number : float
        Number whose precision to determine. Note: if an integer with a 
        floating point (e.g `1.`) is input, it will be considered a 
        float number with precision 1.

    Returns
    -------
    precision : int
        The negative log ofo a float containing the digit 1 in the same 
        position as the last decimal point of the input number. If the number
        does not contain decimal points, then the value 0 is returned.
    
    
    """
    
    if '.' not in str(number):
        return 0
    

    decimals = str(number).split('.')[1]
    i=0
    while i < len(decimals):
        i += 1
        
    return i


def Gauss(t, a, t0, fwhm):
    """
    Returns a Gaussian shape given an array of x value and Gaussian parameters.
    
    Parameters
    ----------
    t : arr
        Array of x values
    a : float
        Amplitude of Gaussian
    t0 : float
        center of Gaussian shape on `t`
    fwhm : float
        Full-width at half-maximum of Gaussian shape
    
    Returns
    -------
    gauss : arr
        Array with same shape as t that follows a Gaussian curve
    """
    
    sigma = fwhm/(2*np.sqrt(2*np.log(2)))
    gauss = a * np.exp(-(t - t0)**2 / (2 * sigma**2))
    return gauss

def ExpGauss(t, t0, a, fwhm, tau, b=0):
    """ 
    Parameters
    ----------
    t : numpy array
        The time array on which to build the IRF
    t0 : float
        Mean of Gaussian profile (before exponential modification)
    a : float
        Amplitude of function 
    fwhm : float
        The FWHM that determines the variance of either type of curve. Must 
        have the same time units as `t`.
    tau : float
        Lifetime of decay of exponentially weighted Gaussian. Ignored if 
        `weighted` is set to False.
    b : float
        Offset of the final function, recommended as zero.
        
    """
    
    k = 1/tau # rate of exponential decay
    sigma = fwhm/(2*np.sqrt(2*np.log(2)))
    arg1 = k/2*(2*t0 + k*sigma**2 - 2*t)
    arg2 = (t0 + k*sigma**2 - t) / (np.sqrt(2)*sigma)
    
    one = np.exp(arg1)
    two = special.erfc(arg2)
    return a * k/2 * one * two + b

def GaussDiff(t, t0, a, fwhm, tau, b, c, weighted, tau_wt):
    """ 
    Creates n Exponentially weighted Gaussian IRF with a diffusion tail
    
    Parameters
    ----------
    t : numpy array
        The time array on which to build the IRF
    irf_type : string
        Type of IRF to be constructed; ```irf_type = 'Gauss'``` will return a 
        normal Gaussian curve, while ```irf_type = 'GaussDiff'``` will return
        an exponentially modified Gaussian with a diffusion tail.
    weighted : Boolean
        Whether to use a pure Gaussian or exponentially weighted Gaussian for 
        the main "peak" of the IRF. Default is True. 
        See en.wikipedia.org/wiki/Exponentially_modified_Gaussian_distribution. 
    fwhm : float
        The FWHM that determines the variance of either type of curve. Must 
        have the same time units as `t`.
    tau_wt : float
        lifetime of decay of exponentially weighted Gaussian. Ignored if 
        `weighted` is set to False.
    tau : float
        lifetime of diffusion tail in the time units of `t`. Ignored if 
        `irf_type` is set to `'Gauss'`.
    b : float
        Relative contribution of diffusion tail to IRF. Ignored if 
        `irf_type` is set to `'Gauss'`.
    c : float
        Offset for IRF, recommended as zero.
    
    """
    if weighted == False:
        arg1 = Gauss(t, 1, t0, fwhm)
    else:
        arg1 = ExpGauss(t, t0, 1, fwhm, tau_wt, b=0)
    arg2 = np.exp(-(t-t0)/tau)
    for i, y in enumerate(arg2):
        if t[i] < t0:
            arg2[i] = 0 
    y = np.zeros(len(arg1))
    arg1 /= np.max(arg1)    
    for i in range(len(y)):
        y[i] = np.max([a*arg1[i], a*b*arg2[i]])
    c *=np.max(y)
    y[y<c] = c
    return y 



import time

import numpy as np

from KinetiKit import sim


__all__ = ['simulate_until_steady', 'simulate', 
'simulate_for_cycles', 'simulate_until_steady',
'refined_simulation']


def simulate(p0, RE_set, t_obj, light_obj):
    """
    Simulates ONE pulse cycle given an initial population, rate equation, 
    time, and excitation parameters.
    """
    
    t = t_obj['array']
    dt = t_obj['dt']    
    subsample = t_obj['subsample']

    # --- Incident Photons --- #
    pulse = light_obj.gen_pulse(t[t <= light_obj.pulse_window], center=0.5*light_obj.pulse_window, stepsize=dt)
    cw = light_obj.cw_power

    # --- Population arrays --- #
    out = np.zeros((RE_set.popnum, t.size // subsample))
    p_current = p0.copy() if p0 is not None else np.zeros(RE_set.shape)

    for i in range(t.size):
        p_previous = p_current.copy()

        photons = cw
        if i < pulse.size:
            photons += pulse[i]
        if i<5:
            pass
            #print(i, p_current, RE_set.rate(p_previous, photons))
        
        p_current += dt * RE_set.rate(p_previous, photons)
        # if i<10:
        #     print('dt', dt/1e-9, 'ns, photons', photons)
        #     print(p_current)
        if i % subsample == 0:
            out[:, i // subsample] = p_current

    return out


def simulate_for_cycles(RE_set, t_obj, light_obj, numcycles=1, p0=None):
    """
    Performs a simulation given pulse and rate equation parameters up to a given
    number of pulse cycles. Notifies when steady state is reached but does not
    stop simulation until numcycles is reached. Designed for testing new simulation
    models for convergence.

    Required Parameters
    ----------
    RE_set : system instance
        requires `rate` method
    t_obj : dict
        Key value combinations from sim.time module
    light_obj : excitation object
        Contains information about the pulsed and CW excitation for the 
        simulated experiment.
    
    Optional Parameters
    ----------
    numcycles : integer
        Number of cycles for which simulation occurs. Default is 1.
    p0 : array 
        Initial population from which to start the simulation. Setting p0 to 
        None defaults to a population of zeros. p0 must have the length of
        `RE_set.popnum`.

    Returns
    ----------
    current : array (2D)
        The final iteration of simulation.  Shape of (npop, ntime).
    converged : bool
        Whether or not the simulation converged within numcycles loops

    """
    converged = False
    current = None
    if p0 is None:
        p0 = np.zeros(RE_set.popnum) 
    else:
        p0 = p0
    notified=False
    for c in range(1, numcycles+1):
        current = simulate(p0, RE_set, t_obj, light_obj)
        converged = isSteady(p0, current[:, -1])
        if converged and not notified:
            print("Reached steady after %i cycles. Continuing simulation"%c)
            notified = True
        p0 = current[:, -1]
    return current, converged

def simulate_until_steady(RE_set, t_obj, light_obj, p0=None, verbose=False):
    """
    Performs a simulation given pulse and rate equation parameters until steady
    state is reached.

    Required Parameters
    ----------
    RE_set : system instance
        requires `rate` method
    t_obj : dict
        Key value combinations from sim.time module
    light_obj : excitation object
        Contains information about the pulsed and CW excitation for the 
        simulated experiment.

    Optional Parameters
    ----------
    p0 : array-like
        initial carrier populations; must have the same length as 
        RE_set.populations. Default is all zeros.
    verbose : boolean
        Whether to print debug-friendly informative messages. Default is False.

    Returns
    ----------
    current : array (2D)
        The final iteration of simulation.  Shape of (npop, ntime).
    converged : bool
        Whether or not the simulation converged within light_obj.numCycles loops

    """
    converged = False
    toofast=False
    current = None
    if p0 is None:
        p0 = np.zeros(RE_set.popnum) 
    else:
        p0 = p0
    for c in range(1, light_obj.numcycles+1):
        current = simulate(p0, RE_set, t_obj, light_obj)
        if (current<0).any():
            if verbose:
                print('Too fast')
            toofast=True
            current = np.zeros(current.shape)-1
            break
        converged = isSteady(p0, current[:, -1], tol=0.001)
        if converged:
            if verbose:
                print("Reached steady after %i cycles"%c)
            break
        p0 = current[:, -1]
    if not converged and not toofast:
        print('Failed to reach steady state after %i cycles'%light_obj.numcycles)
    return current, converged

def refined_simulation(RE_set, t_obj, light_obj, N_coarse=500, doubleSearch=False, verbose=False):
    """
    Performs a simulation with coarse time step until steady state is reached, 
    and then performs a final simulation with a finer time step.

    Required Parameters
    ----------
    RE_set : system instance
        requires `rate` method
    t_obj : dict
        Key value combinations from sim.time module
    light_obj : excitation object
        Contains information about the pulsed and CW excitation for the 
        simulated experiment.
    N_coarse : integer
        Number of time steps per cycle taken during the coarse simulation.
        Default is 500.
        
    Optional Parameters
    -------------------
    doubleSearch : boolean
        Specifies whether the steady state population found by the coarse 
        simulation should be taken as is and only be integrated on once 
        (False), or whether a new search for steady state should be performed
        in the fine time scale (True). Setting `doubleSearch = True` guarantees
        that you will reach the same steady state as you would if you ran the 
        entire simulation at the fine time scale, but is unnecessary if the
        prediction of the steady state population by the coarse simulation is
        accurate. Default is False.
    verbose : boolean
        Whether to print debug-friendly informative messages. Default is False.
    
    Returns
    ----------
    fine_sim : array (2D)
        The final iteration of simulation.  
        Shape of (RE_set.popnum, `N_fine/t_obj['subsample']`).
    converged : bool
        Whether or not the simulation converged within t_obj.numCycles loops
    """
    
    #N_fine = t_obj['N']
    to_coarse = sim.time.update_linear(t_obj, **{'N' : N_coarse})
    tc_start = time.process_time()
    coarse_sim, converged = simulate_until_steady(RE_set, to_coarse, light_obj, verbose=verbose)
    coarse_p0 = coarse_sim.transpose()[np.argmin(coarse_sim[0])]
    tc_end = time.process_time()
    
    if (coarse_p0==-1).any(): # checks if coarse simulation led to "toofast" conditions
        #print('Did not refine simulation')
        return np.zeros((len(coarse_p0), t_obj['N'])), converged
    to_fine = t_obj
    
    tf_start = time.process_time()
    if doubleSearch:
        fine_sim, converged = simulate_until_steady(RE_set, to_fine, light_obj, p0=coarse_p0, verbose=verbose)
    else: 
        fine_sim = simulate(coarse_p0, RE_set, to_fine, light_obj)
    tf_end = time.process_time()
    
    #print('Coarse time (s):', tc_end-tc_start)
    #print('Fine time (s):', tf_end-tf_start)
    
    return fine_sim, converged


# --- Functions assisting integration
def isSteady(previous, current, tol=0.001):
    ids = current != 0
    dif = np.abs(current[ids] - previous[ids]) / current[ids]
    return np.all(dif < tol)




"""
Visualization functions and widget toolkits used to create interactive plots \
that overlay data with kinetic fits. See InteractivePlotting_Mono and InteractivePlotting_Hetero examples.

"""
from ._vizfuncs import *
from ._widgettoolkit import *


import numpy as np
import warnings
from KinetiKit.sim.systems import RateModel
warnings.filterwarnings("ignore", category=RuntimeWarning)
    
class Hetero(RateModel):
    """
    Heterostructure of two adjacent systems, excitons, free carriers, and 
    transfer thereof. No trapping.
    """
    
    class_name = 'Hetero'
    default = {
    'k1_ann': 1,
    'k1_dis': 1,
    'k1_rec': 1,
    'cs1': 1,
    
    'k2_ann': 1,
    'k2_dis': 1,
    'k2_rec': 1,
    'cs2': 1,
    
    'k_xtr': 1,
    'k_etr': 1,
    'k_htr': 1,
    }
    
    populations = ['1x', '1e', '1h', '2x', '2e', '2h']
   
    def rate(self, N, photons):
        n1x, n1e, n1h, n2x, n2e, n2h = N
        n1x_rate = photons*self.cs1 \
            - (self.k1_ann + self.k1_dis + self.k_xtr) * n1x
        n1e_rate = self.k1_dis*n1x - self.k1_rec*n1e*n1h - self.k_etr*n1e
        n1h_rate = self.k1_dis*n1x - self.k1_rec*n1e*n1h - self.k_htr*n1h

        n2x_rate = photons*self.cs2 - (self.k2_ann + self.k2_dis) * n2x \
            + self.k_xtr*n1x
        n2e_rate = self.k2_dis*n2x - self.k2_rec*n2e*n2h + self.k_etr*n1e
        n2h_rate = self.k2_dis*n2x - self.k2_rec*n2e*n2h + self.k_htr*n1h

        return np.array([n1x_rate,
                         n1e_rate,
                         n1h_rate,
                         n2x_rate,
                         n2e_rate,
                         n2h_rate])

    def PLsig(self, N):
        # This produces two PL outputs - one for each part of the heterostructure
        out = np.zeros((2, N.shape[-1]))
        n1x = N[self.populations.index('1x')]
        n1e = N[self.populations.index('1e')]
        n1h = N[self.populations.index('1h')]
        n2x = N[self.populations.index('2x')]
        n2e = N[self.populations.index('2e')]
        n2h = N[self.populations.index('2h')]
        out[0] = self.k1_ann * n1x + self.k1_rec * n1e * n1h
        out[1] = self.k2_ann * n2x + self.k2_rec * n2e * n2h
        return out
    


"""
Methods for reading csv files created by this program.
"""

import sys
import os
import csv

import numpy as np

def csv_to_dict(filename, sourcefolder='', headertype='row'):
    """
    Returns a dictionary with keys and values obtained from a CSV file. Useful 
    for importing fit parameters and X-Y traces from previous fits.
    
    Parameters
    ----------
    filename : string
        Filename of CSV file containing the data and headings to be imported.
    sourcefolder : path (or string)
        Location of CSV file.
    headertype : 'row' or 'column'
        Specifies how the headings are arranged in the CSV file. In a
        '*_params.csv' file, each heading takes up a `row`. In a '*_traces.csv'
        file, each heading takes up a column, with arrays following underneath.
    """
    
    p = os.path.join(sourcefolder, filename)
    
    if headertype == 'row':
        # assumes that the headers are arranged in each row, with a single 
        # value following to the right of each header (e.g. a _params file)
        headers = np.genfromtxt(p, delimiter = ',', unpack=True, usecols=[0],
                                dtype=str) 
        values = np.genfromtxt(p, delimiter = ',', unpack=True, usecols=[1],
                               dtype=float)
    if headertype == 'column':
        # assumes that the headers are arranged in each column, with a single
        # value or an array of values stacked vertically below each header
        with open(p, 'r') as f:
            reader = csv.reader(f, delimiter=',')
            # get header from first row
            headers = next(reader)
            # get all the rows as a list
            data = list(reader)
            # transform data into numpy array
            try:
                values = np.transpose(np.array(data).astype(float))
            except ValueError:
                values = np.transpose(np.array(data).astype(str))
    
    outdict = {}
    
    for i, header in enumerate(headers):
        outdict[header] = values[i]
    
    return outdict



"""
Interconversion between arrays, lists, dictionaries
"""

import numpy as np

def dict_from_list(arglist, dictionary):
    """
    Returns a new dictionary with the same keys as `dictionary` where values
    are set from arglist. 
    
    Assumes that `arglist` has the same length as `dictionary`, and that the
    list items are ordered in the same order that `dictionary.keys()` were 
    defined. Alternatively, `dictionary` can be a dictionary.keys() instance.
    
    Parameters
    ----------
    arglist : array-like
        Must have the same number of elements as `dictionary`.
    dictionary : dictionary, or dictionary.keys()
        The output of this function will be a new dictionary with the same 
        set of keys as `dictionary`.
        
    Raises
    ------
    TypeEerror : If `dictionary` is not of the correct type.
    
    Returns
    -------
    new_dict : dictionary
        A new dictionary with same set of keys as `dictionary` and corresponding
        values set to values of `arglist`.
    """
    import collections
    new_dict = {}
    
    if isinstance(dictionary, dict):
        for i, key in enumerate(list(dictionary.keys())):
            new_dict[key] = arglist[i]
    elif isinstance(dictionary,collections.abc.KeysView):
        for i, key in enumerate(list(dictionary)):
            new_dict[key] = arglist[i]
    else:
        raise TypeError('dictionary must be of type dict or dict_keys.')
        #issue_error
    
    return new_dict

def list_to_array(listorarray):
    if isinstance(listorarray, list):
        return np.array(listorarray)
    elif isinstance(listorarray, np.ndarray):
        return listorarray
    #issue_error

 
def make_2d(array):
    array = list_to_array(array)
    if len(array.shape) ==1:
        return(np.reshape(array, (1,len(array))))
    elif len(array.shape) > 2:
        return np.concatenate(array)
    else:
        return array





#--- Units
# The letter 'u' here is meant to replicate the Greek letter Î¼ (mu)
# If you absolutely hate this notation, alternatives using 'micro' instead of u
# are provided

fs = 1e-15
ps = 1e-12
ns = 1e-9
us = 1e-6
microsec = 1e-6
ms = 1e-3
MHz = 1e6
kHz = 1e3
nW = 1e-9
uW = 1e-6
microWatt = 1e-6
nm = 1e-9
um = 1e-6
micron = 1e-6
mm = 1e-3
cm = 1e-2
eV = 1.602e-19
meV = 1.602e-22

#--- Unit Dictionary (useful as an input on functions)
units = {'fs' : fs,
         'ps' : ps,
         'ns' : ns,
         'us' : us,
         'microsec' : microsec,
         'μs' : microsec,
         'ms' : ms,
         'MHz' : MHz, 
         'kHz' : kHz,
         'nm' : nm,
         'μm' : um,
         'nW' : nW,
         'uW' : uW,
         'μW' : uW,
         'microWatt' : microWatt,
         'micron' : micron,
         'mm' : mm,
         'cm' : cm,
         'eV' : eV,
         'meV': meV,
         'none' : 1,
        }

#--- constants
c = 2.9979e8 # speed of light in m/s
hPlanck = 6.626e-34 # Planck's constant in J s
k_B = 1.3806e-23 # Boltzmann's constant in J/K



"""
Methods and tools used throughout the package, which do not pertain exclusively to data or simulation.
"""
from ._mathfuncs import *
from ._alignment import *
from ._savetools import *
from ._readtools import *
from ._ald import *
from ._displaying import *

__all__ = ['Gauss', 
           'ExpGauss',
           'GaussDiff',
           'align_by_max', 
           'align_by_steep',
           'csv_to_dict',
           'dict_from_list',
           'dict_to_csv',
           'find_nearest',
           'list_to_array',
           'normalized',
           'precision',
           'printparamsexp',
           'roll_by_array_shift',
           'find_baseline',
           'save_fit',
           'traces_to_csv']


#! /usr/bin/env python3

from setuptools import setup, find_packages

setup(
    name="KinetiKit",
	version="0.1.9",
	author="Natalia Spitha",
	author_email="natalia.spitha@gmail.com",
    packages=find_packages(exclude=("tests", "tests.*")),
    python_requires=">=3.6",
    install_requires=[
        "matplotlib>=3.0",
        "numpy>=1.15.0",
        "scipy",
		"ipywidgets",
    ],
    description="Tools for comparing and kinetically simulating time-resolved data",
)



import numpy as np

from KinetiKit.units import MHz, fs, nm, uW
import KinetiKit.kit as kin_kit


__all__ = ['Excitation']


class Excitation():
    """define excitation parameters for simulation"""
    
    pulse_default = {'power': 1*uW,
               'reprate': 80*MHz,
               'fwhm': 100*fs,
               'wavelength': 400*nm,
               'pulse_window': None,
               }
    
    cw_default = {'power': 0,
               'wavelength': 514.5*nm
               }
    
    def __init__(self, pulse={}, cw={}, numcycles=500):
        
        pulse_params = self.pulse_default.copy(); pulse_params.update(pulse)
        self.pulse = pulse_params
        self.pulse_power = self.pulse['power'] 
        self.pulse_energy = self.pulse['power'] / self.pulse['reprate']
        self.pulse_wavelength = self.pulse['wavelength']
        self.pulse_carriers = self.pulse_energy \
            * joules_to_photons(self.pulse_wavelength)
        self.pulse_fwhm = self.pulse['fwhm']
        self.pulse_window = self.pulse['pulse_window']
        if self.pulse_window is None:
            self.pulse_window = 20 * self.pulse_fwhm
        
        cw_params = self.cw_default.copy(); cw_params.update(cw)
        self.cw = cw_params
        self.cw_wavelength = self.cw['wavelength']
        self.cw_power = self.cw['power'] * joules_to_photons(self.cw_wavelength)
        
        self.numcycles = numcycles
        
    def gen_pulse(self, t, center=0, stepsize=None):
        """
        generates a Gaussian pulse along a time array; if the time step is too large
        compared to the FWHM of the pulse, a 1x1 array is generated that contains the 
        rate of arrival of photons.
        """
        pulseCarriers = self.pulse_carriers
        sigma_to_fwhm = 2 * np.sqrt(2 * np.log(2))
        pulsePeak = pulseCarriers * sigma_to_fwhm \
            / (np.sqrt(2 * np.pi) * self.pulse_fwhm)
        if len(t) > 50:
            return kin_kit.Gauss(t, pulsePeak, center, self.pulse_fwhm)
        else:
            return np.array([pulseCarriers/stepsize])
    
    def updated_with(self, pulse={}, cw={}, numcycles=None):
        """returns a new light object based on the current light object, but
        with some modified arguments.
        
        """
        new_pulse = self.pulse; new_pulse.update(pulse)
        new_cw = self.cw; new_cw.update(cw)
        if numcycles is None:
            numcycles = self.numcycles
        return Excitation(new_pulse, new_cw, numcycles)
        

def joules_to_photons(wavelength):
    """convert Joules to photons

    arguments
    wavelength : units of meters
    """
    h_Planck = 6.626e-34  # J s
    c = 2.9979e8  # m / s
    return wavelength / (h_Planck * c)




__all__ = ['simulate_func']

def simulate_func(function, t):
    """
    Simulates a system based on a mathematical function along a time axis.
    
    Required Parameters
    ----------
    function : system instance
        requires `PLsig` method
    t_obj : dict
        Key value combinations from sim.time module
    
    """    
    out = function.PLsig(t)
    
    return out, True
    